/****************************************************************************
*																			*
*							HPACK Multi-System Archiver						*
*							===========================						*
*																			*
*					   		  LZA Main Compressor Code						*
*							  LZA.C  Updated 31/12/91						*
*																			*
*   This program is protected by copyright and as such if you use or copy	*
*   this code for your own purposes directly or indirectly your soul will	*
*   become the property of the author with the right to dispose of it as	*
* 									he wishes.								*
*																			*
*		Copyright 1989 - 1991  Peter C.Gutmann.  All rights reserved		*
*																			*
****************************************************************************/

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "defs.h"
#ifdef SYSV
  #include <stdio.h>
  #include "../choice.h"
  #include "../error.h"
  #include "../hpacklib.h"
  #include "../system.h"
  #include "../crc/crc16.h"
  #include "../crypt/crypt.h"
  #include "../io/fastio.h"
  #include "../io/hpackio.h"
  #include "../lza/model.h"
  #include "../lza/model3.h"
  #include "../lza/model4.h"
#else
  #include "choice.h"
  #include "error.h"
  #include "hpacklib.h"
  #include "crc/crc16.h"
  #include "crypt/crypt.h"
  #include "io/fastio.h"
  #include "io/hpackio.h"
  #include "lza/model.h"
  #include "lza/model3.h"
  #include "lza/model4.h"
#endif /* SYSV */

/* Prototypes for the arithmetic coder */

void startEncode( void );
void endEncode( void );
void encode( const int cumFreq0, const int cumFreqLower, const int cumFreqUpper );

void startDecode( void );
int decodeOrder1( void );
int decodeOrder0( const int *cumFreq );

/* Prototypes for the models for various encodings */

void startLiteralModel( void );
void startHighPosModel( void );
void startLowPosModel( void );

/* Prototypes for the order-1 arithcoder */

void initModels( void );
void endModels( void );
void haveAGoAt( const int value );

/* LZSS Coding Parameters */

#define MAX_POSITION		16384	/* Window size - must be a power of 2 */
#define MAX_LENGTH			64		/* Max. match length - must be pwr.of 2 */

#define STR_BUF_SIZE		MAX_POSITION	/* Size of string buffer */
#define STR_BUF_MASK		( STR_BUF_SIZE - 1 )	/* For fast "% STR_BUF_SIZE" */
#define LOOKAHEAD_BUF_SIZE	MAX_LENGTH		/* Size of look-ahead buffer */
#define LOOKAHEAD_BUF_MASK	( LOOKAHEAD_BUF_SIZE - 1 )	/* For fast "% LOOKAHEAD_BUF_SIZE" */
#define LOOKAHEAD_BUF_START	( STR_BUF_SIZE - LOOKAHEAD_BUF_SIZE )
#define MATCH_THRESHOLD		5	/* When changing this value, remember to change
								   NO_SYMBOLS in crunch/model.h, and the
								   hashTable size in crunch/model.c */

#define LIST_SIZE			8192	/* The number of hashed list headers */
#define LIST_END			-1		/* Marker for end of list */

static BYTE *textBuffer, *lookAheadBuffer;	/* The string and lookahead buffers */
static int *linkBuffer, *tailList;	/* The list1 data structure */
int *headList;						/* Must be externally visible for findMatch() */

/* The output byte count and inFD/outFD (used by PACK.C, UNPACK.C) */

long outByteCount;
FD inFD, outFD;

/* The count of the number of bytes processed for the variable-adaptivity
   count scaling */

unsigned int modelByteCount;

/* The byte count for which we bother updating modelByteCount.  Some
   compilers can't handle 'U' so we declare it as 'L' instead.  If possible
   it should be declared as 'U' for 16-bit systems */

#ifdef IRIX
  #define SCALE_RANGE		40000L
#else
  #define SCALE_RANGE		40000U
#endif /* IRIX */

/****************************************************************************
*																			*
*								Encoding Routines							*
*																			*
****************************************************************************/

WORD isText;	/* Actually a BOOLEAN but WORD makes it faster to handle */

/* Set up the models for the various encoders */

void startModels( void )
	{
	/* Set up the models for the encoder */
	h291();
	startHighPosModel();
	startLowPosModel();

	/* Reset the byte count for the count scaling */
	modelByteCount = 0;
	}

/* Encode a literal */

#define encodeLiteral(value)	haveAGoAt( value )

/* Encode a high position */

void encodeHighPos( const int theCode )
	{
	const int symbol = highPosToIndex[ theCode ];	/* Translate code to an index */

	encode( highPosCumFreq[ 0 ], highPosCumFreq[ symbol - 1 ], highPosCumFreq[ symbol ] );

	/* Now update the model */
	updateHighPosModel( symbol );
	}

/* Encode a low position */

void encodeLowPos( const int theCode )
	{
	const int symbol = lowPosToIndex[ theCode ];	/* Translate code to an index */

	encode( lowPosCumFreq[ 0 ], lowPosCumFreq[ symbol - 1 ], lowPosCumFreq[ symbol ] );

	/* Now update the model */
	if( !isText )
		updateLowPosModel( symbol );
	}

/* Decode a literal */

#define decodeLiteral()	decodeOrder1()

/* Decode a high position */

int decodeHighPos( void )
	{
	const int symbol = decodeOrder0( highPosCumFreq );
	const int retVal = indexToHighPos[ symbol ];

	/* Update the model */
	updateHighPosModel( symbol );

	/* Return translated value */
	return( retVal );
	}

/* Decode a low position */

int decodeLowPos( void )
	{
	const int symbol = decodeOrder0( lowPosCumFreq );
	const int retVal = indexToLowPos[ symbol ];

	/* Update the model */
	if( !isText )
		updateLowPosModel( symbol );

	/* Return translated value */
	return( retVal );
	}

/****************************************************************************
*																			*
*					LZSS Compression/Decompression Routines					*
*																			*
****************************************************************************/

/* Various variables used by the assembly language findMatch() routine */

int listIndex, lookAheadIndex, matchPos, lastMatchLen, currBufPos;
WORD linkBufferBase, textBufferBase, lookAheadBase, headListBase;
long noBytes;

/* Set up buffers for LZSS compression and order-1 arithcoder */

void initPack( const BOOLEAN initAllBuffers )
	{
	if( ( textBuffer = hmalloc( STR_BUF_SIZE ) ) == NULL )
		error( OUT_OF_MEMORY );

	/* Only allocate the following buffers if we need them - saves 64K */
	if( initAllBuffers )
		{
		if( ( lookAheadBuffer = hmalloc( LOOKAHEAD_BUF_SIZE ) ) == NULL ||
			( linkBuffer = hmalloc( ( STR_BUF_SIZE * sizeof( int ) ) ) ) == NULL ||
			( headList = hmalloc( LIST_SIZE * sizeof( int ) ) ) == NULL ||
			( tailList = hmalloc( LIST_SIZE * sizeof( int ) ) ) == NULL )
			error( OUT_OF_MEMORY );
		}

	initModels();
	}

/* Free LZSS buffers and order-1 arithcoder data */

void endPack( void )
	{
	hfree( textBuffer );
	if( choice != EXTRACT && choice != TEST && choice != DISPLAY )
		{
		hfree( lookAheadBuffer );
		hfree( linkBuffer );
		hfree( headList );
		hfree( tailList );
		}
	endModels();
	}

/* Perform the compression */

LONG pack( BOOLEAN *localIsText, const FD inFD, const FD localOutFD, long localNoBytes )
	{
	int ch, count, currMatchLen;
	long printCount;
	BOOLEAN textBufferFull = FALSE, flipFlop = FALSE;
	int textBufferCount = 0, offset = 0, endByte = 0;
	int tailIndex;
	int headIndex;
	int srcIndex, destIndex;
	BYTE *srcPtr;

	/* Copy various local paramater to global paramters */
	outFD = localOutFD;
	noBytes = localNoBytes;

	/* Evaluate size of input file */
	noBytes += CHECKSUM_LEN;
	printCount = noBytes - 2048L;	/* Make sure we don't print a spurious dit */

	/* Initialise the I/O system */
	resetFastIn( inFD );

	/* Initialise the arithmetic encoder and models */
	startModels();
	startEncode();
	outByteCount = 0L;

	/* See if the file is a text file or binary and use the appropriate
	   model.  If at least 95% of the file's chars are in the following
	   category, we assume it's text */
	for( count = 0, currMatchLen = 0; count < 50; count++ )
		if( ( ( ( ( ch = _inBuffer[ count ] ) >= 0x20 ) && ( ch <= 0x7E ) ) ) || \
			ch == '\r' || ch == '\n' || ch == '\t' )
			currMatchLen++;
	*localIsText = isText = ( currMatchLen > 47 ) ? TRUE : FALSE;

	/* Initialize various data structures */
	memset( headList, LIST_END, LIST_SIZE * sizeof( int ) );
	memset( tailList, LIST_END, LIST_SIZE * sizeof( int ) );
	memset( linkBuffer, LIST_END, STR_BUF_SIZE * sizeof( int ) );
	currBufPos = 0;

	/* Now fill the lookahead buffer */
	for( lookAheadIndex = 0; \
		 lookAheadIndex < LOOKAHEAD_BUF_SIZE && ( ch = fgetByte( inFD ) ) != FEOF; \
		 lookAheadBuffer[ lookAheadIndex++ ] = ch );
	lookAheadIndex = 0;

	do
		{
		/* Find the longest match of the lookahead buffer in the text buffer */
		lastMatchLen = 1;
		listIndex = lookAheadBuffer[ lookAheadIndex ] << 5;
		ch = lookAheadBuffer[ ( lookAheadIndex + 1 ) & LOOKAHEAD_BUF_MASK ];
		ch ^= lookAheadBuffer[ ( lookAheadIndex + 2 ) & LOOKAHEAD_BUF_MASK ];
		listIndex |= ch & 0x1F;
		headIndex = headList[ listIndex ];
		while( headIndex != LIST_END )
			{
			/* Match ( position + 1 ) (since we already know there is a match
			   at ( position )) with the text in the lookahead buffer.  Note
			   that it is cheaper to recalculate destIndex inside the loop
			   every iteration than to precalculate it externally */
			destIndex = ( lookAheadIndex + 1 ) & LOOKAHEAD_BUF_MASK;
			currMatchLen = 1;

			/* Make sure we don't try and match past the current position in
			   the buffer.  Use an offset of -1 since we've already matched
			   the first char */
			count = currBufPos - headIndex - 1;
			if( count < 0 )
				count += STR_BUF_SIZE;
			if( count >= LOOKAHEAD_BUF_SIZE )
				count = LOOKAHEAD_BUF_SIZE - 1;

			if( headIndex < STR_BUF_SIZE - LOOKAHEAD_BUF_SIZE )
				{
				/* Do fast match without wraparound check */
				srcPtr = textBuffer + headIndex + 1;
				while( ( *srcPtr++ == lookAheadBuffer[ destIndex ] ) &&
					   count-- )
					{
					currMatchLen++;
					destIndex = ++destIndex & LOOKAHEAD_BUF_MASK;
					}
				}
			else
				{
				/* Slow match with checking for wraparound */
				srcIndex = ( headIndex + 1 ) & STR_BUF_MASK;
				while( ( textBuffer[ srcIndex++ ] == \
						 lookAheadBuffer[ destIndex ] ) && count-- )
					{
					srcIndex &= STR_BUF_MASK;
					destIndex = ++destIndex & LOOKAHEAD_BUF_MASK;
					currMatchLen++;
					}
				}

			/* See if the new match is longer than an older match */
			if( currMatchLen >= lastMatchLen )
				{
				lastMatchLen = currMatchLen;
				matchPos = headIndex;
				}

			/* Move to next match in list */
			headIndex = linkBuffer[ headIndex ];
			}

		/* Special case:  At the very end of the file we may have a match
		   longer than the number of chars left.  In this case we decrease
		   the match length, and compress as single chars if necessary */
		if( noBytes < lastMatchLen )
			lastMatchLen = ( int ) noBytes;

		/* See if we have a match long enough to make it worthwhile ouputting
		   it as an LZSS code */
		if( lastMatchLen <= MATCH_THRESHOLD )
			{
			/* Output literal character */
			lastMatchLen = 1;
			encodeLiteral( lookAheadBuffer[ lookAheadIndex ] );
			}
		else
			{
			/* Output ( length, position ) pair */
			encodeLiteral( ( lastMatchLen + 256 ) - MATCH_THRESHOLD );
			if( isText )
				/* Take position from end of match not start */
				matchPos += lastMatchLen;
			offset = ( currBufPos - matchPos ) & STR_BUF_MASK;
			encodeHighPos( offset >> 7 );
			encodeLowPos( offset & 0x7F );
			}

		/* Now move the matched string from the lookahead buffer to the text
		   buffer, updating the head and tail lists as we go */
		for( count = 0; count < lastMatchLen; count++ )
			{
			/* See if we need to start deleting chars from the text buffer */
			if( textBufferFull )
				{
				/* Unlink the old char at currBufPos from lists */
				listIndex = textBuffer[ currBufPos ] << 5;
				ch = textBuffer[ ( currBufPos + 1 ) & STR_BUF_MASK ];
				ch ^= textBuffer[ ( currBufPos + 2 ) & STR_BUF_MASK ];
				listIndex |= ch & 0x1F;
				tailIndex = tailList[ listIndex ];
				if( tailIndex == currBufPos )
					/* Last char deleted */
					headList[ listIndex ] = tailList[ listIndex ] = LIST_END;
				else
					/* Point to next char in list */
					headList[ listIndex ] = linkBuffer[ currBufPos ];
				}
			else
				{
				textBufferCount++;

				/* Check whether the buffer is now full */
				if( textBufferCount >= STR_BUF_SIZE )
					textBufferFull = TRUE;
				}

			/* Move a char from the lookahead buffer into the string buffer
			   and read in a new char to replace it */
			textBuffer[ currBufPos ] = ch = lookAheadBuffer[ lookAheadIndex ];
			if( noBytes <= CHECKSUM_LEN + LOOKAHEAD_BUF_SIZE )
				lookAheadBuffer[ lookAheadIndex ] = ( isText ) ? \
									CHECKSUM_TEXT[ endByte++ ] : \
									CHECKSUM_BIN[ endByte++ ];
			else
				lookAheadBuffer[ lookAheadIndex ] = fgetByte( inFD );
			lookAheadIndex = ++lookAheadIndex & LOOKAHEAD_BUF_MASK;

			/* Add it to the link buffer/lists at currBufPos */
			listIndex = ch << 5;
			ch = lookAheadBuffer[ lookAheadIndex ]; /* lAI has been inc'd in the meantime */
			ch ^= lookAheadBuffer[ ( lookAheadIndex + 1 ) & LOOKAHEAD_BUF_MASK ];
			listIndex |= ch & 0x1F;
			tailIndex = tailList[ listIndex ];
			if( tailIndex == LIST_END )
				/* No chars in list */
				headList[ listIndex ] = currBufPos;
			else
				/* Add to end of list */
				linkBuffer[ tailIndex ] = currBufPos;
			tailList[ listIndex ] = currBufPos;
			linkBuffer[ currBufPos ] = LIST_END;

			/* Increment currBufPos */
			currBufPos = ++currBufPos & STR_BUF_MASK;

			/* Decrement bytes seen */
			noBytes--;
			if( modelByteCount < SCALE_RANGE )
				modelByteCount++;
			}

		/* Output a dit for every 4K processed */
		if( noBytes < printCount )
			{
			if( flipFlop )
				{
				hputchars( '\b' );
				hputchars( 'O' );
				}
			else
				hputchars( 'o' );
			hflush( stdout );
			flipFlop = !flipFlop;
			printCount -= 2048L;
			}
		}
	while( noBytes > 0 );

	/* Clean up any loose ends */
	endEncode();
	if( flipFlop )
		{
		hputchars( '\b' );
		hputchars( 'O' );
		hflush( stdout );
		}

	return( outByteCount );
	}

/* Perform the decompression */

LONG dataLen;	/* Used in UNPACK.C */

BOOLEAN unpack( const BOOLEAN localIsText, const FD localInFD, const FD localOutFD, \
				const LONG dataSize, const long localNoBytes )
	{
	int position, length, bufIndex, ch;
	int endByte = 0, printCount = 0, checkBytes = 0;
	BYTE checkSum[ CHECKSUM_LEN ];
	BOOLEAN flipFlop = FALSE;

	/* Copy the local parameters to their global equivalents */
	inFD = localInFD;
	outFD = localOutFD;
	noBytes = localNoBytes;

	/* Initialise the I/O system */
	dataLen = dataSize;
	resetFastOut();
	noBytes += CHECKSUM_LEN;	/* Non-test version stores only orig.file size */

	/* Determine which model was used for compression */
	isText = localIsText;

	/* Initialise various data structures */
	startModels();
	startDecode();
	memset( checkSum, 0xFF, CHECKSUM_LEN );	/* Clear checksum */

	bufIndex = 0;
	while( noBytes > 0 )
		{
		/* Decode a literal char or length byte */
		ch = decodeLiteral();
		if( ch < 256 )
			{
			/* Output byte and add to circular buffer */
			if( noBytes <= CHECKSUM_LEN )
				{
				/* If the data is corrupted we may end up decoding a random
				   string, so we insert a sanity check here */
				if( endByte >= CHECKSUM_LEN )
					return( FALSE );

				checkSum[ endByte++ ] = ch;
				}
			else
				fputByte( outFD, ch );

			textBuffer[ bufIndex++ ] = ch;
			bufIndex &= STR_BUF_MASK;
			noBytes--;
			printCount++;
			if( modelByteCount < SCALE_RANGE )
				modelByteCount++;
			}
		else
			{
			/* Decode ( length, position ) pair */
			length = ( ch - 256 ) + MATCH_THRESHOLD;
			position = decodeHighPos() << 7;
			position |= decodeLowPos();
			position = bufIndex - position;
			if( isText )
				position -= length;
			position &= STR_BUF_MASK;

			/* Adjust byte count and printCount for string */
			noBytes -= length;
			if( noBytes <= CHECKSUM_LEN )
				checkBytes = CHECKSUM_LEN - ( int ) noBytes;
			printCount += length;
			if( modelByteCount < SCALE_RANGE )
				modelByteCount += length;

			/* Copy length bytes, starting at position, to the output */
			while( length-- )
				{
				/* Read bytes of the string out of the circular buffer */
				ch = textBuffer[ position++ & STR_BUF_MASK ];
				if( checkBytes && length < checkBytes )
					{
					/* If the data is corrupted we may end up decoding a
					   random string, so we insert a sanity check here */
					if( endByte >= CHECKSUM_LEN )
						return( FALSE );

					checkSum[ endByte++ ] = ch;
					}
				else
					fputByte( outFD, ch );

				/* Add them to the tail end of the buffer */
				textBuffer[ bufIndex++ ] = ch;
				bufIndex &= STR_BUF_MASK;
				}
			}

		/* Output a dit for every 4K processed */
		if( printCount >= 2048 )
			{
			if( flipFlop )
				{
				hputchars( '\b' );
				hputchars( 'O' );
				}
			else
				hputchars( 'o' );
			hflush( stdout );
			flipFlop = !flipFlop;
			printCount -= 2048;
			}
		}

	/* Flush the output buffer and print the last dit if necessary */
	flush( outFD );
	if( flipFlop )
		{
		hputchars( '\b' );
		hputchars( 'O' );
		hflush( stdout );
		}

	/* Compare the final CHECKSUM_LEN decoded bytes against the checksum as
	   a data integrity check */
	return( !memcmp( checkSum, ( isText ) ? CHECKSUM_TEXT : CHECKSUM_BIN, \
					 CHECKSUM_LEN ) );
	}
